from sympy import *
import re
import multiprocessing
import os
import math

class Logic_expression():
    def __init__(self, logic_expression_string):
        self.logic_expression_string = logic_expression_string
        self._get_symbolic_expression()
        self._get_symbols()
        self._order_symbols()
        self._get_dnf()

    def _get_symbolic_expression(self):
        self.logic_expression_symbolic = parse_expr(self.logic_expression_string, evaluate = False)
    
    def _order_symbols(self):
        symbols_list_string = list(self.used_symbols.keys())
        symbols_list_string.sort()
        symbols_lib = {}
        for symbol_string in symbols_list_string:
            symbols_lib[symbol_string] = self.used_symbols[symbol_string]
        self.used_symbols = symbols_lib
        self.symbols_list = list(symbols_lib.values())
        
    def _get_symbols(self):
        symbol_lib = {}
        symbols_list = list(self.logic_expression_symbolic.atoms())
        for symbol in symbols_list:
            symbol_lib[str(symbol)] = symbols(str(symbol))
        self.used_symbols = symbol_lib

    
    def _get_truthtable(self):
        self.my_truthtable = Truth_table(self.logic_expression_symbolic, self.used_symbols)
        truthtable_full = self.my_truthtable.truth_table
        self.truthtable = {key:value for (key,value) in truthtable_full.items() if value == True}

    def _get_dnf_string(self):
        dnf_string = []
        true_terms = self.truthtable.keys()
        for true_term in true_terms:
            dnf_string.append(self._get_term_string(true_term))
        self.dnf_string = '|'.join(dnf_string)

    def _get_term_string(self, term):
        term_bin = ('{0:0'+str(len(self.used_symbols))+'b}').format(term)
        term_string = []
        for i in range(0,len(term_bin)):
            if int(term_bin[i]):
                term_string.append(str(self.symbols_list[i]))
            else:
                term_string.append('~'+ str(self.symbols_list[i]))
        return '('+ '&'.join(term_string) + ')'

    def _get_dnf(self):
        self._get_truthtable()
        self._get_dnf_string()

class Truth_table():
    def __init__(self, logic_expression_symbolic, symbols):
        self.logic_expression_symbolic = logic_expression_symbolic
        self.symbols = symbols
        self.symbols_string = list(self.symbols.keys())
        self._get_truth_table()
    
    def _create_table_entry(self, binary_i):
        symbol_dict = {}
        for i in range(0,len(self.symbols_string)):
            symbol_dict[self.symbols_string[i]] = binary_i[i]
        return symbol_dict

    def _evaluate_entry(self, tt_entry):
        return self.logic_expression_symbolic.subs(tt_entry)
        
    def _get_truth_table(self):
        truth_table_size = 2 ** (len(self.symbols))
        self.truth_table = {}
        for i in range(0,truth_table_size):
            binary_i = ('{0:0'+str(len(self.symbols))+'b}').format(i)
            tt_entry = self._create_table_entry(binary_i)
            self.truth_table[i] = self._evaluate_entry(tt_entry)

class Logic_expression_list():
    def __init__(self, full_expression_list):
        self.full_expression_list = full_expression_list
        self._process_parallel()

    def _expression_list_worker(self, part_expression_list,queue=None):
        result = {}
        for expression in part_expression_list:
            expression_instance = Logic_expression(expression)
            result[expression] = expression_instance.dnf_string
        if queue == None:
            return result
        else:
            queue.put(result)

    def _process_parallel(self):
        cpu_count = os.cpu_count()
        chunks_count = int(math.ceil(len(self.full_expression_list) / cpu_count))
        queue = multiprocessing.Queue()
        procs = []
        procs_range = range(0,min(cpu_count,len(self.full_expression_list)))
        for i in procs_range:
            proc = multiprocessing.Process(target=self._expression_list_worker, args=(self.full_expression_list[chunks_count*i:chunks_count*(i+1)], queue))
            procs.append(proc)
            proc.start()
        
        results ={}
        for i in procs_range:
            results.update(queue.get())
        
        for i in procs:
            i.join()
        self.results = results
def get_dnfs(string_List):
    my_Instance = Logic_expression_list(string_List)
    return(my_Instance.results)


from flask import Flask, url_for, request, render_template, redirect, make_response
from werkzeug.utils import secure_filename
import os
import dnf_converter as dnf

upload_files = "/home/uni/EP/logic_normalizer/upload_files" #Order für die hochgeladenen Dateien
extensions = set(['txt']) #Nur txt Dateien erlauben
app = Flask(__name__)

def allowed(filename):
    return '.' in filename and filename.rsplit('.', 1)[1].lower() in extensions #1.check: enthält die Datei einen Punkt 2.Check ist der rechte Teil hinter dem letzten Punkt in den Extentions

def read_file(file):
    file_content = open(os.path.join(upload_files, file.filename),'r')
    return file_content.read().split('\n')


@app.route("/", methods=["GET","POST"])
def index():
    if request.method == "POST":
        if 'file' not in request.files:
            logic_expression = request.form["logic_expression"]
            result = dnf.get_dnfs([logic_expression])
            return render_template('index.html', result=result)
            
        file = request.files['file']
        if allowed(file.filename):
            filename = secure_filename(file.filename)
            file.save(os.path.join(upload_files, filename))
            file_content = read_file(file)
            result = dnf.get_dnfs(file_content)
            print(result)
            return render_template('index.html', result = result)

    else:
        return render_template('index.html', result = {})

def start_flask():
    app.run(port=1337, debug = True)
if __name__ == '__main__':
    app.run(port=1337, debug = True)
    


import unittest as ut
from dnf_converter.cli import main
from dnf_converter.dnf_converter import *
from sympy import *
import random

class TestDnfConverter(ut.TestCase):
    def test_main(self):
        assert main([]) == 0
    
    def test_Logic_expression(self):
        test_expression_string = '(A|B)&C'
        test_expression = Logic_expression(test_expression_string)
        assert test_expression.dnf_string == '(~A&B&C)|(A&~B&C)|(A&B&C)'

    def test_get_symbols_List(self):
        test_expression_string = '(A|B)&C'
        test_expression = Logic_expression(test_expression_string)
        assert test_expression.symbols_list == [symbols('A'),symbols('B'),symbols('C')]

    def test_get_symbols_symbols(self):
        test_expression_string = '(A|B)&C'
        test_expression = Logic_expression(test_expression_string)
        test_lib ={'A':symbols('A'),'B':symbols('B'),'C':symbols('C')}
        assert test_expression.used_symbols == test_lib

    def test_get_truthtable(self):
        test_expression_string = '(A|B)&C'
        test_expression = Logic_expression(test_expression_string)
        assert test_expression.truthtable == {3:True,5:True,7:True}

    def test_Truth_table(self):
        test_expression_string = '(A|B)&C'
        logic_expression_symbolic = parse_expr(test_expression_string, evaluate = False)
        symbol_dict = {'A':symbols('A'),'B':symbols('B'),'C':symbols('C')}
        test_truth_table = Truth_table(logic_expression_symbolic, symbol_dict)
        assert test_truth_table.truth_table == {0:False, 1:False, 2:False, 3:True, 4:False, 5:True, 6:False, 7:True}

    def test_create_table_entry(self):
        test_expression_string = '(A|B)&C'
        logic_expression_symbolic = parse_expr(test_expression_string, evaluate = False)
        symbol_dict = {'A':symbols('A'),'B':symbols('B'),'C':symbols('C')}
        test_truth_table = Truth_table(logic_expression_symbolic, symbol_dict)
        assert test_truth_table._create_table_entry('000') == {'A':'0','B':'0','C':'0'}
        assert test_truth_table._create_table_entry('010') == {'A':'0','B':'1','C':'0'}

    def test_evaluate_entry(self):
        test_expression_string = '(A|B)&C'
        logic_expression_symbolic = parse_expr(test_expression_string, evaluate = False)
        symbol_dict = {'A':symbols('A'),'B':symbols('B'),'C':symbols('C')}
        test_truth_table = Truth_table(logic_expression_symbolic, symbol_dict)
        assert test_truth_table._evaluate_entry({'A':'0','B':'0','C':'0'}) == False
        assert test_truth_table._evaluate_entry({'A':'1','B':'1','C':'1'}) == True

    def test_expression_list_worker(self):
        test_list = ['(A|B)&C','A&B','~A|B']
        worker = Logic_expression_list(test_list)
        result_list = worker._expression_list_worker(test_list)
        assert result_list == {'(A|B)&C': '(~A&B&C)|(A&~B&C)|(A&B&C)', 'A&B': '(A&B)', '~A|B': '(~A&~B)|(~A&B)|(A&B)'}

    def test_process_parallel(self):
        full_expression_list = ['(A|B)&C','A&B','~A|B']
        expr_List = Logic_expression_list(full_expression_list)
        assert expr_List.results == {'(A|B)&C': '(~A&B&C)|(A&~B&C)|(A&B&C)', 'A&B': '(A&B)', '~A|B': '(~A&~B)|(~A&B)|(A&B)'}
    
    def test_Masstest(self):
        number_of_teststrings = 20
        max_number_of_symbols = 10
        my_teststring_set = Teststringset(number_of_teststrings, max_number_of_symbols)
        my_teststrings = my_teststring_set.string_set
        expr_List = Logic_expression_list(my_teststrings)
        initial_string_List = list(expr_List.results.keys())
        for i in range(number_of_teststrings):
            string1 =  initial_string_List[i]
            string2 =  expr_List.results[string1]
            if len(string2):
                combined_string = '~' +'(' + string1  + ')'+ '&' +'(' + string2 + ')'
            else:
                continue
            print(combined_string)
            sub_dict = {}
            parsed_expression = parse_expr(combined_string)
            for sym in parsed_expression.atoms():
                sub_dict[sym] = random.randint(0,1)
            print(i)
            assert parsed_expression.subs(sub_dict) == False


class Teststringset():
    def __init__(self, number_of_teststrings, max_number_of_symbols):
        self.number_of_teststrings = number_of_teststrings
        self.max_number_of_symbols = random.randint(1,max_number_of_symbols)
        self._create_symbols()
        self._create_strings()

    def _create_symbols(self):
        self.symbols = []
        for i in range(0,self.max_number_of_symbols):
            self.symbols.append('sym'+str(i))

    def _create_strings(self):
        self.string_set = []
        for i in range(0,self.number_of_teststrings):
            self.available_symbols = self.max_number_of_symbols
            new_string = self._create_string()
            self.string_set.append(new_string)

    def _create_string(self, bracket_mode = False):
        expression_string = ''
        while self.available_symbols > 0:
            if random.randint(0,1): ##zufaellige not einfuegen
                expression_string = expression_string +'~'
            if not random.randint(0,3): ##Entscheigung ob neuer geklammerter ausdruck oder lediglich ein symbol eingesetzt werden soll
                bracket_string = self._create_string(bracket_mode=True)
                expression_string = expression_string + '(' + bracket_string + ')'
            else:
                expression_string = expression_string + random.choice(self.symbols) ##zufaellige variable auswaehlen
                self.available_symbols = self.available_symbols - 1 ##anzahl der verfuegbaren variablen verringern
            if random.randint(0,1): ## 'und' oder 'oder' einfuegen 
                expression_string = expression_string +'|'
            else:
                expression_string = expression_string +'&'

            if bracket_mode == True and not random.randint(0,2):
                break

        expression_string = expression_string[0:-1]
        return expression_string


if __name__=='__main__':
    ut.main()


"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mdnf_converter` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``dnf_converter.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``dnf_converter.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
import sys
import argparse
import dnf_converter as dnf
import argparse
import .flask_app as fa
##from flask_app import *

def main(argv=sys.argv):
  parser=argparse.ArgumentParser()
  parser.add_argument('--web', help='Start a Webserver')
  parser.add_argument('--cli', help='stay in the commandline')
  args=parser.parse_args()
  print(args)
  if args.web is not None:
    fa.start_flask()
    return 0
  elif args.cli is not None:
    result = dnf.get_dnfs(args.cli)
    return result
  else:
    return 0
  

if __name__ == '__main__':
  main()